#include "tracer.h"

#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <linux/unistd.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <time.h>
#include <unistd.h>

#include "rt-utils.h"

/*
 * From cyclictest code source
 */

#define KVARS 32
#define KVARNAMELEN 32
#define KVALUELEN 32

static char *fileprefix;
static char *procfileprefix = "/proc/sys/kernel/";
static char *debugfileprefix = "/sys/kernel/debug/tracing/";
static int trace_fd = -1;

static int kernvar(int mode, const char *name, char *value,
                   size_t sizeofvalue) {
  char filename[128];
  int retval = 1;
  int path;
  size_t len_prefix = strlen(fileprefix), len_name = strlen(name);

  memcpy(filename, fileprefix, len_prefix);
  memcpy(filename + len_prefix, name, len_name + 1);

  path = open(filename, mode);
  if (path >= 0) {
    if (mode == O_RDONLY) {
      int got;
      if ((got = read(path, value, sizeofvalue)) > 0) {
        retval = 0;
        value[got - 1] = '\0';
      }
    } else if (mode == O_WRONLY) {
      if (write(path, value, sizeofvalue) == sizeofvalue) retval = 0;
    }
    close(path);
  }
  return retval;
}

static void setkernvar(const char *name, char *value) {
  if (kernvar(O_WRONLY, name, value, strlen(value)))
    fprintf(stderr, "could not set %s to %s\n", name, value);
}

static int trace_file_exists(char *name) {
  struct stat sbuf;
  char *tracing_prefix = get_debugfileprefix();
  char path[MAX_PATH];
  strcat(strcpy(path, tracing_prefix), name);
  return stat(path, &sbuf) ? 0 : 1;
}

static int settracer(char *tracer) {
  if (valid_tracer(tracer)) {
    setkernvar("current_tracer", tracer);
    return 0;
  }
  return -1;
}

void setup_tracer(void) {

  fileprefix = procfileprefix;
  setkernvar("ftrace_enabled", "1");

  fileprefix = debugfileprefix;
  // Clear old traces by setting tracer to nop first
  settracer("nop");
  settracer("function");

  /* open the tracing on file descriptor */
  if (trace_fd == -1) {
    char path[MAX_PATH];
    strcpy(path, fileprefix);
    if (trace_file_exists("tracing_on"))
      strcat(path, "tracing_on");
    else
      strcat(path, "tracing_enabled");
    if ((trace_fd = open(path, O_WRONLY)) == -1)
      printf("unable to open %s for tracing", path);
  }

  tracing(0);
}

void tracing(int on) {
  if (on)
    write(trace_fd, "1", 1);
  else
    write(trace_fd, "0", 1);
}
